/* 
 * SMTP processor for sslwrap
 *
 * Written by Rick Kaseguma
 * Copyright 1999 Rick Kaseguma
 * http://www.rickk.com/sslwrap/
 *
 * September 1, 1999
 *
 */
#include <ctype.h>  /* tolower */
#include <stdarg.h> /* va_list */
#include <string.h> /* memmove */

#include "smtp.h"

int smtpInit(SmtpState *pState, int sockIn, int sockOut, SSL *pConn)
{
  pState->state = SMTP_INIT;
  pState->sockIn = sockIn; /* this is s_in, where we read unencrypted data from the local server from */
  pState->sockOut = sockOut; /* this is s_out, where we write unencrypted data to the local server to */
  pState->pConn = pConn /* this is the encrypted socket */;
  pState->serverIsESMTP = 0;

  pState->debug = 0; /* fopen( "/tmp/ssldebug", "a" ); */ 
  
  smtpDebug( pState, "%x starting\n", pState);

  return 0; /* success */
}
/* Called after getting data from the encryped side, before writing to the
 * local host unencrypted data 
 */

int smtpIn(SmtpState *pState, char *pBuf, int bufLen)
{
  char cmdBuf[20];
  int i;

  if (pState->state != SMTP_DATA) {
    for(i = 0; (i < (sizeof(cmdBuf) - 2)) && (i < bufLen) && (pBuf[i] != '\015'); i++) {
      cmdBuf[i] = tolower( pBuf[i] );
    }
    cmdBuf[i] = 0;
    smtpDebug( pState, "%x probable command '%s'\n", pState, cmdBuf);
  }

  switch( pState->state) {
  case SMTP_INIT:
    /* Error: Client shouldn't send before banner.
     */
    smtpDebug( pState, "%x client sent commands before banner!\n", pState );
    return 0;

  case SMTP_GOTBANNER:
    /* We are waiting to initialize the connection. We are looking for a
     * EHLO from the client
     */
    if (bufLen < 4) {
      /* error... should deal with partial buffer sends but this will be 
       * ok for now
       */
      smtpSend( pState, "501 Bad parameters for authenticated login\n" );
      smtpDebug( pState, "%x client sent short response to banner (%d)\n", pState, bufLen );
      pState->state = SMTP_IGNORE;
      return 0;
    }
    if (strncmp(cmdBuf, "ehlo", 4) == 0) {
      if (pState->serverIsESMTP) {
	/* is an ESMTP server, so just pass the data along */
	smtpDebug( pState, "%x server is ESMTP, got EHLO from client\n", pState );
	pState->state = SMTP_GOTEHLO;
      } else {
	/* server is not ESMTP, so we need to fake an ESMTP reply */
	smtpDebug( pState, "%x server is not ESMTP, faking\n", pState);
	pState->state = SMTP_GOTEHLO;
	strncpy( pBuf, "HELO", 4);
      }
    } else {
      /* not EHLO... reject it! */
      smtpDebug( pState, "%x client didn't EHLO!\n", pState );
      smtpSend( pState, "501 Authenticated login using ESMTP is required!\015\012");
      pState->state = SMTP_IGNORE;
      return 0;
    }
    break;
  
  case SMTP_AUTHWAIT:
    if (strncmp(cmdBuf, "auth login", 10) == 0) {
      /* this is what we want -- auth login */
      smtpDebug( pState, "%x client did auth login\n", pState);
      smtpSend( pState, "334 VXNlcm5hbWU6\015\012");
      bufLen = 0; /* intercept data from client */
      pState->state = SMTP_USERWAIT;
    } else if (strncmp( cmdBuf, "auth", 4) == 0) {
      smtpDebug( pState, "%x client did auth, but not login!\n", pState);
      smtpSend( pState, "501 LOGIN is the only AUTH flavor supported\015\012" );
      bufLen = 0; /* intercept data from client */
      pState->state = SMTP_IGNORE;
    } else {
      /* bad - */
      smtpDebug( pState, "%x client sent a command other than AUTH!\n", pState);
      smtpSend( pState, "501 Authenticated login using AUTH is required\015\012" );
      bufLen = 0; /* intercept data from client */
      pState->state = SMTP_IGNORE;
    }
    break; 

  case SMTP_USERWAIT:
    smtpDebug( pState, "%x client sent username len %d test 3\n", pState, bufLen );

    smtpDecodeBase64( pBuf, bufLen < 16 ? bufLen : 16, cmdBuf );
    strcpy( pState->user, cmdBuf );

    smtpDebug( pState, "%x decoded username '%s'\n", pState, cmdBuf );
    
    smtpSend( pState, "334 UGFzc3dvcmQ6\015\012" );
    pState->state = SMTP_PASSWAIT;
    bufLen = 0; /* intercept data from client */
    break;

  case SMTP_PASSWAIT:
    smtpDebug( pState, "%x client sent password\n", pState );

    smtpDecodeBase64( pBuf, bufLen, cmdBuf );
    /* smtpDebug( pState, "%x decoded password '%s'\n", pState, cmdBuf ); */

    if (!password_verify("sslwrap", pState->user, cmdBuf)) {
      smtpSend( pState, "235 Authentication successful\015\012" );
      smtpDebug( pState, "%x going into data state\n", pState );
      pState->state = SMTP_DATA;
    } else {
      smtpDebug( pState, "%x invalid user or pass\n", pState );
      smtpSend( pState, "501 Invalid username or password\015\012" );
      pState->state = SMTP_IGNORE;
    }

    bufLen = 0; /* intercept data from client */
    break;

  case SMTP_DATA:
    /* let data through */
    break;

  default:
    return 0;
  }
  return bufLen;
}

/* Called after reading data from the unencrypted side, before writing to
 * the encrypted socket 
 */
int smtpOut(SmtpState *pState, char *pBuf, int bufLen)
{
  char *cp;
  switch( pState->state) {
  case SMTP_INIT:
    /* We are waiting to initialize the connection 
     * If the server is doing old smtp, fake it to be ESMTP 
     */
    smtpDebug( pState, "%x got banner from server\n", pState);
    pBuf[bufLen] = 0;
    smtpDebug( pState, "%s\n", pBuf );
    
    cp = strstr( pBuf, " SMTP");
    if (cp) {
      memmove( &cp[1], cp, bufLen - (cp - pBuf));
      cp[1] = 'E';
      bufLen++;
    } else {
      pState->serverIsESMTP = 1;
    }
    pState->state = SMTP_GOTBANNER;
    break;
  case SMTP_GOTEHLO: 
    /* Slip in the "250-AUTH=LOGIN\015\012" line */
    pBuf[bufLen] = 0;
    smtpDebug( pState, "%x got HELO/EHLO response from server\n", pState);
    smtpDebug( pState, "Before: %s\n", pBuf );

    cp = strstr( &pBuf[3], "250");
    if (!cp) {
      cp = pBuf;
    }
    memmove( &cp[16], cp, bufLen - (cp - &pBuf[3]));
    memmove( cp, "250-AUTH=LOGIN\015\012", 16 );
    bufLen += 16;

    pBuf[bufLen] = 0;
    smtpDebug( pState, "After: %s\n", pBuf );

    pState->state = SMTP_AUTHWAIT;
    break;

  case SMTP_USERWAIT:
  case SMTP_PASSWAIT:
    smtpDebug(pState, "%x Odd! Got data from local server in state %d\n", pState, pState->state );
    break;

  case SMTP_DATA: 
    break;

  case SMTP_IGNORE:
    bufLen = 0;
    break;

  default:
    break;
  }
  return bufLen;
}


int smtpTerm(SmtpState *pState)
{
  smtpDebug( pState, "%x stopping\n", pState);
  if (pState->debug){
    fclose(pState->debug);
  }

  return 0; /* success */
}

int smtpSend(SmtpState *pState, const char *str)
{
  int off = 0, len = strlen(str), res;
  while(off < len) {
    res = SSL_write( pState->pConn, &str[off], len);
    if (res > 0) {
      off += res;
    } else if (! BIO_sock_should_retry(res) ) {
      break;
    }
  }
  return len;
}

int smtpDebug(SmtpState *pState, const char *fmt, ... )
{
  va_list ap;

  if (pState->debug) {
    va_start(ap, fmt );
    
    vfprintf( pState->debug, fmt, ap );
    
    fflush( pState->debug );

    va_end( ap );
  }
}

int smtpDecodeBase64(const char *bufPtr, int bufLen, char *dstPtr)
{
  unsigned char tmp[10], c;
  unsigned int i, tmpi = 0;

  *dstPtr = 0;

  for(i = 0; i < bufLen; i++) {
    c = (unsigned char) bufPtr[i];
    if (c >= 'A' && c <= 'Z') {
      tmp[tmpi++] = c - 'A';
    } else if (c >= 'a' && c <= 'z') {
      tmp[tmpi++] = c - 'a' + 26;
    } else if (c >= '0' && c <= '9') {
      tmp[tmpi++] = c - '0' + 52;
    } else if (c == '+') {
      tmp[tmpi++] = 62;
    } else if (c == '/') {
      tmp[tmpi++] = 63;
    } else if (c == '=') {
      /* pad */
      if (tmpi <= 1) tmp[1] = 0;
      if (tmpi <= 2) tmp[2] = 0;
      if (tmpi <= 3) tmp[3] = 0;
    } else{
      /* invalid character */
      break;
    }

    if (tmpi >= 4) {
      *dstPtr++ = (unsigned char)((tmp[0] << 2) | ((tmp[1] >> 4) & 0x3));
      *dstPtr++ = (unsigned char)((tmp[1] << 4) | ((tmp[2] >> 2) & 0xF));
      *dstPtr++ = (unsigned char)((tmp[2] << 6) | (tmp[3] & 0x3F)); 
      tmpi = 0;
    }
  }

  if (tmpi >= 2) {
      *dstPtr++ = (unsigned char)((tmp[0] << 2) | ((tmp[1] >> 4) & 0x3));
  }
  if (tmpi >= 3) {
      *dstPtr++ = (unsigned char)((tmp[1] << 4) | ((tmp[2] >> 2) & 0xF));
  }
  *dstPtr = 0;

  return 0; 
}
